
from kivy.animation import AnimationTransition
from math import pi, cos, sin

class Transformations():

	@staticmethod
	def bouncey(points, alpha):
		x = AnimationTransition.out_elastic(alpha)
		x0, y0, x1, y1 = points
		h = (y1 - y0) * x  # tend to correct height, from 0
		w = (x1 - x0) * (2 - x)

		return [
			x0 + w * (1 - alpha * alpha), y0,
			(x0 + w * (1 - alpha * alpha) + w), y0,
			(x0 + w * (1 - alpha * alpha) + w), (y0 + h),
			x0 + w * (1 - alpha * alpha), (y0 + h)
		]

	@staticmethod
	def sky_down(points, alpha):
		x = AnimationTransition.out_quad(alpha)
		x0, y0, x1, y1 = points
		h = (y1 - y0) * (4 - 3 * x)  # tend to correct height, from 0
		w = (x1 - x0) * (x)
		cx = (x0 + x1) / 2

		return [
			(cx - w * 0.5), y0,
			(cx + w * 0.5), y0,
			(cx + w * 0.5), (y0 + h),
			(cx - w * 0.5), (y0 + h),
		]

	@staticmethod
	def pop_in(points, alpha):
		x = AnimationTransition.out_elastic(alpha)
		x0, y0, x1, y1 = points
		h = (y1 - y0) * x  # tend to correct height, from 0
		w = (x1 - x0) * x
		cx = (x0 + x1) / 2

		return [
			(cx - w * 0.5), y0,
			(cx + w * 0.5), y0,
			(cx + w * 0.5), (y0 + h),
			(cx - w * 0.5), (y0 + h),
		]

	@staticmethod
	def comes_and_go(points, alpha):
		x = AnimationTransition.in_out_quad(alpha)
		x0, y0, x1, y1 = points

		h = (y1 - y0) * (1 - 2 * abs(x - .5))
		center_y = (y0 + y1) / 2

		w = (x1 - x0) * (1 - 2 * abs(x - .5))
		center_x = (x0 + x1) / 2

		return [
			(center_x - w / 2), (center_y - h / 2),
			(center_x + w / 2), (center_y - h / 2),
			(center_x + w / 2), (center_y + h / 2),
			(center_x - w / 2), (center_y + h / 2)
		]

	@staticmethod
	def roll_in(points, alpha):
		x0, y0, x1, y1 = points
		x = AnimationTransition.out_quad(alpha)
		size = [ (x1 - x0), (y1 - y0) ]
		size_1 = [ (size[0] * x / 2), (size[1] * x / 2) ]

		cx = (x0 + x1) / 2 + size[1] * (1 - x)
		cy = (y0 + y1) / 2 - size[1] * 0.4 * (1 - x)
		a = (pi * x)
		pi4 = (pi / 4)

		return [
			(cx + cos(a + 1 * pi4) * size_1[0]), (cy + sin(a + 1 * pi4) * size_1[1]),
			(cx + cos(a + 3 * pi4) * size_1[0]), (cy + sin(a + 3 * pi4) * size_1[1]),
			(cx + cos(a + 5 * pi4) * size_1[0]), (cy + sin(a + 5 * pi4) * size_1[1]),
			(cx + cos(a + 7 * pi4) * size_1[0]), (cy + sin(a + 7 * pi4) * size_1[1]),
		]





from kivy.lang import Builder
from kivy.core.image import Image as CoreImage
from kivy.properties import (
    NumericProperty, ListProperty,
    BoundedNumericProperty, StringProperty,
    ObjectProperty,
)
from kivy.animation import Animation
from kivy.clock import Clock
from kivy.properties import BooleanProperty
from kivy.uix.widget import Widget

Builder.load_string('''

<ProgressSpinnerBase>:
	_size: min(self.height, self.width)
	_rsize: self._size / 2.0
	_stroke: max(0.1, self._rsize / 20.0 if self.stroke_width == None else self.stroke_width)
	_radius: self._rsize - self._stroke * 2.0

<ProgressSpinner>:
	canvas:
		Color:
			rgba: self.color
		Line:
			circle:
				(self.center_x, self.center_y, self._radius,
				self._angle_center + self._angle_start,
				self._angle_center + self._angle_end)
			width: self._stroke
			cap: 'none'

<TextureProgressSpinner>:
	canvas:
		StencilPush
		Color:
			rgba: [1, 1, 1, 1]
		Line:
			circle:
				(self.center_x, self.center_y, self._radius,
				self._angle_center + self._angle_start,
				self._angle_center + self._angle_end)
			width: self._stroke
			cap: 'none'
		StencilUse

		Color:
			rgba: self.color
		Rectangle:
			pos: self.center_x - self._rsize, self.center_y - self._rsize
			size: self._size, self._size
			texture: self.texture

		StencilUnUse
		Color:
			rgba: [1, 1, 1, 1]
		Line:
			circle:
				(self.center_x, self.center_y, self._radius,
				self._angle_center + self._angle_start,
				self._angle_center + self._angle_end)
			width: self._stroke
			cap: 'none'
		StencilPop

<RotatingTextureProgressSpinner>:
	canvas:
		PushMatrix
		Rotate:
			angle: -(self._angle_center)
			origin: self.center
		
		StencilPush
		Color:
			rgba: [1, 1, 1, 1]
		Line:
			circle:
				(self.center_x, self.center_y, self._radius,
				self._angle_start, self._angle_end)
			width: self._stroke
			cap: 'none'
		StencilUse

		Color:
			rgba: self.color
		Rectangle:
			pos: [ (self.center_x - self._rsize), (self.center_y - self._rsize) ]
			size: [self._size, self._size]
			texture: self.texture

		StencilUnUse
		Color:
			rgba: [1, 1, 1, 1]
		Line:
			circle:
				(self.center_x, self.center_y, self._radius,
				self._angle_start, self._angle_end)
			width: self._stroke
			cap: 'none'
		StencilPop
		
		PopMatrix

''')


class ProgressSpinnerBase(Widget):

    color = ListProperty([1, 1, 1, 1])
    '''Color to render the spinner.
    '''

    speed = BoundedNumericProperty(1, min=0.1)
    '''Speed coefficient of the spinner. This value is multiplied by the
    base speed of 90 degrees per second.
    '''

    stroke_length = BoundedNumericProperty(25., min=1, max=180)
    '''Base length of the stroke in degrees.
    '''

    stroke_width = NumericProperty(None, allownone=True)
    '''Width of the stroke in pixels. If set to None, the width will be
    calculated automatically as 1/20th of the radius.
    '''

    auto_start = BooleanProperty(True)
    '''Whether to automatically start spinning.
    '''

    # internal properties
    _angle_center = NumericProperty(0)
    _angle_start = NumericProperty(0)
    _angle_end = NumericProperty(0)
    _size = NumericProperty(0)
    _rsize = NumericProperty(0)
    _stroke = NumericProperty(1)
    _radius = NumericProperty(50)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._state = 'wait1'
        self._next = None
        self._spinning = False

        if self.auto_start:
            self.start_spinning()

    def start_spinning(self, *args):
        '''Start spinning the progress spinner. Ignores all positional args
        for easy binding.
        '''
        if not self._spinning:
            self._state = 'wait1'
            self._next = None
            self._angle_center = 0.
            self._angle_start = 360.
            self._angle_end = 360. + self.stroke_length
            Clock.schedule_interval(self._update, 0)
            Clock.schedule_once(self._rotate, 0.3)
            self._spinning = True

    def stop_spinning(self, *args):
        '''Stop spinning the progress spinner. Ignores all positional args
        for easy binding.

        If you intend to keep the spinner around, you should stop it when
        not using it and restart it when needed again.
        '''
        if self._spinning:
            if self._next:
                if isinstance(self._next, Animation):
                    self._next.cancel(self)
                else:
                    self._next.cancel()
            Clock.unschedule(self._update)
            Clock.unschedule(self._rotate)
            self._angle_start = self._angle_end = 0
            self._spinning = False

    def _update(self, dt):
        angle_speed = 90. * self.speed
        self._angle_center += dt * angle_speed

        if self._angle_center > 360:
            self._angle_center -= 360.

    def _rotate(self, *args):
        if not self._spinning:
            return None

        rotate_speed = 0.6 / self.speed
        wait_speed = 0.3 / self.speed
        if self._state == 'wait1':
            self._state = 'rotate1'
            self._next = Animation(
                _angle_end=(self._angle_start + 360.0 - self.stroke_length),
                d=rotate_speed, t='in_quad'
            )
            self._next.bind(on_complete=self._rotate)
            self._next.start(self)

        elif self._state == 'rotate1':
            self._state = 'wait2'
            self._next = Clock.schedule_once(self._rotate, wait_speed)

        elif self._state == 'wait2':
            self._state = 'rotate2'
            self._next = Animation(_angle_start=self._angle_end - self.stroke_length, d=rotate_speed, t='in_quad')
            self._next.bind(on_complete=self._rotate)
            self._next.start(self)

        elif self._state == 'rotate2':
            self._state = 'wait1'
            self._next = Clock.schedule_once(self._rotate, wait_speed)
            while self._angle_end > 720.:
                self._angle_start -= 360.
                self._angle_end -= 360.


class ProgressSpinner(ProgressSpinnerBase):
    pass


class TextureProgressSpinnerBase(ProgressSpinnerBase):
    texture = ObjectProperty(None)
    '''Texture to render for the spinner.
    '''

    source = StringProperty('')
    '''Source image to render for the spinner.
    '''

    def on_source(self, inst, value):
        if value:
            self.texture = CoreImage(value).texture


class TextureProgressSpinner(TextureProgressSpinnerBase):
    '''Same as ProgressSpinner, but with a texture/image instead of a solid color.
    '''
    pass


class RotatingTextureProgressSpinner(TextureProgressSpinnerBase):
    '''Same as TextureProgressSpinner, but the texture/image rotates along with the spinner.
    '''
    pass




from kivy.metrics import dp
from kivy.clock import Clock
from kivy.uix.gridlayout import GridLayout
from kivy.properties import ListProperty
from kivy.core.window import Window
from kivygo.behaviors.hover import HoverBehavior
from kivy.properties import ListProperty, NumericProperty
from kivygo.behaviors.drag_and_drop import DraggableLayoutBehavior
from kivy.lang.builder import Builder


Builder.load_string("""

<ColoredGridLayout>:
	canvas.before: 
		Color:
			rgba: self.background_color
		RoundedRectangle:
			pos: self.pos
			size: self.size
			radius: self.radius
		Color:
			rgba: [1, 1, 1, 1]
	canvas.after:
		Color:
			rgba: self.stroke_color
		Line:
			rounded_rectangle: [*self.pos, *self.size, *self.radius]
			width: self.stroke_width
		Color:
			rgba: [1, 1, 1, 1]

""")


class ColoredGridLayout(GridLayout, HoverBehavior):
	background_color = ListProperty([0, 0, 0, 0])
	radius = ListProperty([0, 0, 0, 0])
	stroke_color = ListProperty([0, 0 ,0 ,0])
	stroke_width = NumericProperty(dp(2))


class AdjustSizeGridLayout(ColoredGridLayout):

	max_size = ListProperty([dp(235), dp(250)])

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.config)
	
	def config(self, *args):
		Window.bind(on_flip=self.set_grid_cols)
		self.set_grid_cols()


	def on_size(self, *args):
		Clock.schedule_once(self.set_grid_cols)

	def set_grid_cols(self, *args):
		w1, w2 = self.max_size
		tc = max(2, len(self.children)+1)

		for i in range(1, tc):
			w = round(self.width/i)

			if (w > w1 and w < w2) or w < w1:

				self.cols =  i - (1 if w < w1 and i > 1 else 0)
				break

			elif i == (tc - 1):
				
				self.cols = i
				break
		
		self.padding = 0
		self.spacing = 0
		Clock.schedule_once(self.check_max_width)
	
	def check_max_width(self, *args):
		if not self.children:
			return None

		w1, w2 = self.max_size

		if self.cols >= len(self.children):

			if self.children[0].width > (w2+dp(20)):

				t = int( (self.children[0].width - w2) / 2 )
				self.padding = [t, 0, t, 0]
				self.spacing = t
				return None
		
		self.padding = dp(10)
		self.spacing = dp(10)


class DraggableGridLayout(ColoredGridLayout, DraggableLayoutBehavior):

	def compare_pos_to_widget(self, widget, pos):
		x, y = pos

		if y > widget.top:
			return 'before'

		elif y < widget.y:
			return 'after'

		elif x > widget.right:
			return 'after'

		elif x < widget.x:
			return 'before'

		else:
			spacer = self.spacer_widget
			if widget.parent is spacer.parent:

				children = widget.parent.children
				if children.index(spacer) > children.index(widget):
					return 'after'

		return 'before'




from math import sin, cos, atan2, radians
from kivy.uix.widget import Widget
from kivy.graphics import Color, SmoothLine, Ellipse
from kivy.properties import NumericProperty, ColorProperty, AliasProperty


class RadialSlider(Widget):
    '''A :class:`~kivy.uix.widget.Widget` that provides a radial slider.
    '''

    angle = NumericProperty(0)
    '''
    Current angle used for the Radial Slider.
    You can use it for setting the angle of the thumb in the track in range
    0 - 360.
    '''

    track_color = ColorProperty('#ffffff')
    '''Color of the track.
    '''

    thumb_color = ColorProperty('#ffffff')
    '''Color of the thumb.
    '''

    thumb_diameter = NumericProperty(10)
    '''Diameter of the thumb.
    '''

    track_thickness = NumericProperty(2)
    '''Thickness of the track.
    '''

    max_value = NumericProperty(100)
    '''Maximum value allowed for the Radial Slider.
    '''

    min_value = NumericProperty(0)
    '''Minimum value allowed for the Radial Slider.
    '''

    def get_value(self):
        value_range = abs(self.min_value - self.max_value)
        return self.min_value + max(0, min(self.angle, 360))/360 * value_range

    def set_value(self, value):
        self.angle = max(0, min((value - self.min_value)/abs(
            self.max_value - self.min_value) * 360, 360))

    value = AliasProperty(get_value, set_value,
                          bind=('angle', 'min_value', 'max_value'),
                          cache=True)
    '''
    Normalized value inside the :attr:`range` (min_value/max_value).
    You can use it for setting the value betwwen the minimum and maximum value.
    '''

    def get_norm_value(self):
        return self.angle/360

    def set_norm_value(self, value):
        self.angle = value * 360

    value_normalized = AliasProperty(get_norm_value, set_norm_value,
                                     bind=('value', 'min_value', 'max_value',
                                           'angle'),
                                     cache=True)
    '''
    Normalized value inside the :attr:`range` (min_value/max_value) to
    0-1 range. You can also use it for setting the real value without
    knowing the minimum and maximum value.
    '''

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._start_angle = 0
        self._end_angle = 360
        self._touch_control = None

        with self.canvas:
            self._track_color = Color(1, 1, 1, 1)
            self._track = SmoothLine(width=2, overdraw_width=2.5,
                                     circle=(self.center_x, self.center_y,
                                             0.5*self.width, 0, 360, 800))
        with self.canvas.after:
            self._thumb_color = Color(1, 1, 1, 1)
            self._thumb = Ellipse(size=self.size, pos=self.pos)
            self._thumb_border = SmoothLine(width=1,
                                            overdraw_width=(2.5),
                                            circle=(self._thumb.pos[0],
                                                    self._thumb.pos[1],
                                                    (self.thumb_diameter/2),
                                                    0, 360, 800))

        self.bind(
            size=self.update_canvas,
            pos=self.update_canvas,
            angle=self.update_canvas
        )

    def update_canvas(self, instance, value):
        self._track_color.rgba = self.track_color
        self._track.width = self.track_thickness

        if self.track_thickness < self.thumb_diameter:
            track_adjustment = self.thumb_diameter
        else:
            track_adjustment = self.track_thickness

        min_dimension = min(self.width/2, self.height/2)

        self._track.circle = (self.center_x, self.center_y, min_dimension -
                              track_adjustment, self._start_angle,
                              self._end_angle, 800)

        self._thumb_color.rgba = self.thumb_color
        self._thumb.size = self.thumb_diameter * 2, self.thumb_diameter * 2

        angle = 90 - max(self._start_angle, min(self.angle, self._end_angle))
        spacing = self.track_thickness - self.thumb_diameter
        spacing = 0 if spacing < 0 else spacing
        self._thumb.pos = [
            self.center_x + (min_dimension - self.thumb_diameter) * \
                cos(radians(angle)) - self.thumb_diameter - spacing * \
                    cos(radians(angle)),
            
            self.center_y + (min_dimension - self.thumb_diameter) * \
                sin(radians(angle)) - self.thumb_diameter - spacing * \
                    sin(radians(angle))
        ]

        self._thumb_border.circle = [
            (self._thumb.pos[0] + self.thumb_diameter),
            (self._thumb.pos[1] + self.thumb_diameter),
            self.thumb_diameter, 0, 360, 800,
        ]

    def on_touch_down(self, touch):
        if self.disabled or self._touch_control != None:
            return None
        
        if not self.collide_point(*touch.pos):
            return None

        # check if the touch is in the dragable thumb
        thumb_limits = (touch.x >= self._thumb.pos[0],
                        touch.x <= self._thumb.pos[0] + self._thumb.size[0],
                        touch.y >= self._thumb.pos[1],
                        touch.y <= self._thumb.pos[1] + self._thumb.size[1]
        )
        if not all(thumb_limits):
            return None

        touch.grab(self)
        self._touch_control = touch
        return True

    def on_touch_move(self, touch):
        if touch.grab_current is not self:
            return None
        
        angle = self.get_angle(touch.x, touch.y) * 180 / (3.14)

        if self._start_angle != 0 or self._end_angle != 360:
            self.angle = self.get_angle_norm(angle)
        else:
            # avoid overscrolling to 0
            if (touch.x + 1 > self.center_x) and self.angle > 270:
                self.angle = 360
            # avoid overscrolling to 360
            elif (touch.x - 1 <= self.center_x) and self.angle < 90:
                self.angle = 0
            else:
                self.angle = self.get_angle_norm(angle)

        return True

    def on_touch_up(self, touch):
        if touch.grab_current is not self:
            return None
        
        touch.ungrab(self)
        self._touch_control = None
        return True

    def get_angle(self, x, y):
        cx, cy = self.center
        delta_x = x - cx
        delta_y = y - cy
        return atan2(delta_y, delta_x)

    # current angle in range 1 - 360
    def get_angle_norm(self, angle):
        if round(angle) >= 90 and round(angle) <= 180:
            return round(360 + (90 - angle))
        else:
            return round(90 - angle)




from kivy.uix.layout import Layout
from kivy.properties import BoundedNumericProperty, ListProperty


class NotEnoughCellsException(Exception):
    pass


class SimpleTableLayout(Layout):

    cols = BoundedNumericProperty(None, min=0, allownone=True)
    rows = BoundedNumericProperty(None, min=0, allownone=True)
    _grid = ListProperty([])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bind(
            children=self._trigger_layout,
            size=self._trigger_layout,
            pos=self._trigger_layout
        )

    def add_widget(self, widget):
        if not hasattr(widget, 'colspan'):
            widget.colspan = 1  # TODO warning, not defined as property.

        if not hasattr(widget, 'rowspan'):
            widget.rowspan = 1  # TODO warning, not defined as property.

        return super().add_widget(widget)

    def on_children(self, *args):
        children_cells = sum([c.rowspan * c.colspan for c in self.children])
        total_cells = self.cols * \
            self.rows if self.cols != None and self.rows != None else None

        if total_cells and children_cells > total_cells:
            raise NotEnoughCellsException(f"Available cells: {total_cells}. Requested cells: {children_cells}. Increase cols and/or rows")

    def do_layout(self, *largs):
        if not self.children:
            return None

        cols_width = self.width / self.cols
        rows_height = self.height / self.rows

        grid = [[0 for x in range(self.cols)] for y in range(self.rows)]
        # A grid with size cols x rows.
        # each position represents a cell. zero means the cell is available,
        
        for c in self.children[::-1]:
            c.size = cols_width * c.colspan, rows_height * c.rowspan
            # Find next available cell
            cur_row, cur_col = self._next_cell(grid)
            if cur_row == None or cur_col == None:  # TODO raise exception?
                break

            # fill cell or cells in grid according to rowspan, colspan
            for ry in range(c.rowspan):
                # if rowspan > 1 we need to put the widget at the lowest row
                last_row = cur_row + ry
                for rx in range(c.colspan):
                    grid[cur_row + ry][cur_col + rx] = c
            
            c.x = (self.x + cols_width * cur_col)
            c.y = (self.y + self.height - rows_height - rows_height * last_row)

        self._grid = grid

    def cell(self, row, col):
        """Returns widget at pos (row, col) in the grid"""
        return self._grid[row - 1][col - 1]

    def _next_cell(self, grid):
        # TODO optimize
        for y in range(len(grid)):
            for x in range(len(grid[0])):
                if grid[y][x] == 0:
                    return y, x
        
        return [None, None]





from kivy.animation import Animation
from kivy.uix.stencilview import StencilView
from kivy.metrics import dp
from kivy.clock import Clock
from kivy.properties import (
	ObjectProperty, NumericProperty,
	OptionProperty, BooleanProperty,
	StringProperty,
)
from kivy.lang import Builder


Builder.load_string('''

<NavigationDrawer>:
	size_hint: [1, 1]
	_side_panel: sidepanel
	_main_panel: mainpanel
	_join_image: joinimage
	side_panel_width: min(dp(250), 0.5*self.width)
	BoxLayout:
		id: sidepanel
		y: root.y
		x: root.x - \
		   (1-root._anim_progress)* \
		   root.side_panel_init_offset*root.side_panel_width
		height: root.height
		width: root.side_panel_width
		opacity: root.side_panel_opacity + \
				 (1-root.side_panel_opacity)*root._anim_progress
		canvas:
			Color:
				rgba: [0, 0, 0, 1]
			Rectangle:
				pos: self.pos
				size: self.size
		canvas.after:
			Color:
				rgba: [0, 0, 0, (1-root._anim_progress)*root.side_panel_darkness]
			Rectangle:
				size: self.size
				pos: self.pos
	BoxLayout:
		id: mainpanel
		x: root.x + \
		   root._anim_progress * \
		   root.side_panel_width * \
		   root.main_panel_final_offset
		y: root.y
		size: root.size
		canvas:
			Color:
				rgba: [0, 0, 0, 1]
			Rectangle:
				pos: self.pos
				size: self.size
		canvas.after:
			Color:
				rgba: [0, 0,0 , root._anim_progress*root.main_panel_darkness]
			Rectangle:
				size: self.size
				pos: self.pos
	Image:
		id: joinimage
		opacity: min(sidepanel.opacity, 0 if root._anim_progress < 0.00001 \
				 else min(root._anim_progress*40, 1))
		source: root._choose_image(root._main_above, root.separator_image)
		mipmap: False
		width: root.separator_image_width
		height: root._side_panel.height
		x: (mainpanel.x - self.width + 1) if root._main_above \
		   else (sidepanel.x + sidepanel.width - 1)
		y: root.y
		allow_stretch: True
		keep_ratio: False
''')


class NavigationDrawerException(Exception):
	'''Raised when add_widget or remove_widget called incorrectly on a
	NavigationDrawer.

	'''


class NavigationDrawer(StencilView):
	'''Widget taking two children, a side panel and a main panel,
	displaying them in a way that replicates the popular Android
	functionality. See module documentation for more info.

	'''

	# Internal references for side, main and image widgets
	_side_panel = ObjectProperty(None)
	_main_panel = ObjectProperty(None)
	_join_image = ObjectProperty(None)

	side_panel = ObjectProperty(None, allownone=True)
	'''Automatically bound to whatever widget is added as the hidden panel.'''
	main_panel = ObjectProperty(None, allownone=True)
	'''Automatically bound to whatever widget is added as the main panel.'''

	# Appearance properties
	side_panel_width = NumericProperty()
	'''The width of the hidden side panel. Defaults to the minimum of
	250dp or half the NavigationDrawer width.'''
	separator_image = StringProperty('')
	'''The path to an image that will be placed between the side and main
	panels. If set to `''`, defaults to a gradient from black to
	transparent in an appropriate direction (left->right if side panel
	above main, right->left if main panel on top).'''
	separator_image_width = NumericProperty(dp(10))
	'''The width of the separator image. Defaults to 10dp'''

	# Touch properties
	touch_accept_width = NumericProperty('14dp')
	'''Distance from the left of the NavigationDrawer in which to grab the
	touch and allow revealing of the hidden panel.'''
	_touch = ObjectProperty(None, allownone=True)  # The currently active touch

	# Animation properties
	state = OptionProperty('closed', options=('open', 'closed'))
	'''Specifies the state of the widget. Must be one of 'open' or
	'closed'. Setting its value automatically jumps to the relevant state,
	or users may use the anim_to_state() method to animate the
	transition.'''
	anim_time = NumericProperty(0.3)
	'''The time taken for the panel to slide to the open/closed state when
	released or manually animated with anim_to_state.'''
	min_dist_to_open = NumericProperty(0.7)
	'''Must be between 0 and 1. Specifies the fraction of the hidden panel
	width beyond which the NavigationDrawer will relax to open state when
	released. Defaults to 0.7.'''
	_anim_progress = NumericProperty(0)  # Internal state controlling
										 # widget positions
	_anim_init_progress = NumericProperty(0)

	# Animation controls
	top_panel = OptionProperty('main', options=['main', 'side'])
	'''Denotes which panel should be drawn on top of the other. Must be
	one of 'main' or 'side'. Defaults to 'main'.'''
	_main_above = BooleanProperty(True)

	side_panel_init_offset = NumericProperty(0.5)
	'''Intial offset (to the left of the widget) of the side panel, in
	units of its total width. Opening the panel moves it smoothly to its
	final position at the left of the screen.'''

	side_panel_darkness = NumericProperty(0.8)
	'''Controls the fade-to-black of the side panel in its hidden
	state. Must be between 0 (no fading) and 1 (fades to totally
	black).'''

	side_panel_opacity = NumericProperty(1)
	'''Controls the opacity of the side panel in its hidden state. Must be
	between 0 (fade to transparent) and 1 (no transparency)'''

	main_panel_final_offset = NumericProperty(1)
	'''Final offset (to the right of the normal position) of the main
	panel, in units of the side panel width.'''

	main_panel_darkness = NumericProperty(0)
	'''Controls the fade-to-black of the main panel when the side panel is
	in its hidden state. Must be between 0 (no fading) and 1 (fades to
	totally black).
	'''

	opening_transition = StringProperty('out_cubic')
	'''The name of the animation transition type to use when animating to
	an open state. Defaults to 'out_cubic'.'''

	closing_transition = StringProperty('in_cubic')
	'''The name of the animation transition type to use when animating to
	a closed state. Defaults to 'out_cubic'.'''

	anim_type = OptionProperty('reveal_from_below',
							   options=['slide_above_anim',
										'slide_above_simple',
										'fade_in',
										'reveal_below_anim',
										'reveal_below_simple',
										])
	'''The default animation type to use. Several options are available,
	modifying all possibly animation properties including darkness,
	opacity, movement and draw height. Users may also (and are
	encouaged to) edit these properties individually, for a vastly
	larger range of possible animations. Defaults to reveal_below_anim.
	'''

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		Clock.schedule_once(self.on__main_above, 0)

	def on_anim_type(self, *args):
		anim_type = self.anim_type
		if anim_type == 'slide_above_anim':
			self.top_panel = 'side'
			self.side_panel_darkness = 0
			self.side_panel_opacity = 1
			self.main_panel_final_offset = 0.5
			self.main_panel_darkness = 0.5
			self.side_panel_init_offset = 1
		if anim_type == 'slide_above_simple':
			self.top_panel = 'side'
			self.side_panel_darkness = 0
			self.side_panel_opacity = 1
			self.main_panel_final_offset = 0
			self.main_panel_darkness = 0
			self.side_panel_init_offset = 1
		elif anim_type == 'fade_in':
			self.top_panel = 'side'
			self.side_panel_darkness = 0
			self.side_panel_opacity = 0
			self.main_panel_final_offset = 0
			self.main_panel_darkness = 0
			self.side_panel_init_offset = 0.5
		elif anim_type == 'reveal_below_anim':
			self.top_panel = 'main'
			self.side_panel_darkness = 0.8
			self.side_panel_opacity = 1
			self.main_panel_final_offset = 1
			self.main_panel_darkness = 0
			self.side_panel_init_offset = 0.5
		elif anim_type == 'reveal_below_simple':
			self.top_panel = 'main'
			self.side_panel_darkness = 0
			self.side_panel_opacity = 1
			self.main_panel_final_offset = 1
			self.main_panel_darkness = 0
			self.side_panel_init_offset = 0

	def on_top_panel(self, *args):
		if self.top_panel == 'main':
			self._main_above = True
		else:
			self._main_above = False

	def on__main_above(self, *args):
		newval = self._main_above
		main_panel = self._main_panel
		side_panel = self._side_panel
		self.canvas.remove(main_panel.canvas)
		self.canvas.remove(side_panel.canvas)
		if newval:
			self.canvas.insert(0, main_panel.canvas)
			self.canvas.insert(0, side_panel.canvas)
		else:
			self.canvas.insert(0, side_panel.canvas)
			self.canvas.insert(0, main_panel.canvas)

	def toggle_main_above(self, *args):
		if self._main_above:
			self._main_above = False
		else:
			self._main_above = True

	def add_widget(self, widget):
		if len(self.children) == 0:
			super().add_widget(widget)
			self._side_panel = widget

		elif len(self.children) == 1:
			super().add_widget(widget)
			self._main_panel = widget

		elif len(self.children) == 2:
			super().add_widget(widget)
			self._join_image = widget

		elif self.side_panel == None:
			self._side_panel.add_widget(widget)
			self.side_panel = widget

		elif self.main_panel == None:
			self._main_panel.add_widget(widget)
			self.main_panel = widget
		else:
			raise NavigationDrawerException(
				'Can\'t add more than two widgets'
				'directly to NavigationDrawer')

	def remove_widget(self, widget):
		if widget is self.side_panel:
			self._side_panel.remove_widget(widget)
			self.side_panel = None
		elif widget is self.main_panel:
			self._main_panel.remove_widget(widget)
			self.main_panel = None
		else:
			raise NavigationDrawerException(
				'Widget is neither the side or main panel, can\'t remove it.')

	def set_side_panel(self, widget):
		'''Removes any existing side panel widgets, and replaces them with the
		argument `widget`.
		'''
		# Clear existing side panel entries
		if len(self._side_panel.children) > 0:
			for child in self._side_panel.children:
				self._side_panel.remove(child)
		# Set new side panel
		self._side_panel.add_widget(widget)
		self.side_panel = widget

	def set_main_panel(self, widget):
		'''Removes any existing main panel widgets, and replaces them with the
		argument `widget`.
		'''
		# Clear existing side panel entries
		if len(self._main_panel.children) > 0:
			for child in self._main_panel.children:
				self._main_panel.remove_widget(child)
		# Set new side panel
		self._main_panel.add_widget(widget)
		self.main_panel = widget

	def on__anim_progress(self, *args):
		if self._anim_progress > 1:
			self._anim_progress = 1
		elif self._anim_progress < 0:
			self._anim_progress = 0

		if self._anim_progress >= 1:
			self.state = 'open'
		elif self._anim_progress <= 0:
			self.state = 'closed'

	def on_state(self, *args):
		Animation.cancel_all(self)
		self._anim_progress = (1) if self.state == 'open' else (0)
		
	def anim_to_state(self, state):
		'''If not already in state `state`, animates smoothly to it, taking
		the time given by self.anim_time. State may be either 'open'
		or 'closed'.

		'''
		if state == 'open':
			anim = Animation(_anim_progress=1,
							 duration=self.anim_time,
							 t=self.closing_transition)
			anim.start(self)

		elif state == 'closed':
			anim = Animation(_anim_progress=0,
							 duration=self.anim_time,
							 t=self.opening_transition)
			anim.start(self)
		
		else:
			raise NavigationDrawerException(
				'Invalid state received, should be one of `open` or `closed`')

	def toggle_state(self, animate=True):
		'''Toggles from open to closed or vice versa, optionally animating or
		simply jumping.'''
		if self.state == 'open':
			if animate:
				self.anim_to_state('closed')
			else:
				self.state = 'closed'
		elif self.state == 'closed':
			if animate:
				self.anim_to_state('open')
			else:
				self.state = 'open'

	def on_touch_down(self, touch):
		col_self = self.collide_point(*touch.pos)
		col_side = self._side_panel.collide_point(*touch.pos)
		col_main = self._main_panel.collide_point(*touch.pos)

		if self._anim_progress < 0.001:  # i.e. closed
			valid_region = (self.x <=
							touch.x <=
							(self.x + self.touch_accept_width))
			if not valid_region:
				self._main_panel.on_touch_down(touch)
				return False
		else:
			if col_side and not self._main_above:
				self._side_panel.on_touch_down(touch)
				return False
			valid_region = (self._main_panel.x <=
							touch.x <=
							(self._main_panel.x + self._main_panel.width))
			if not valid_region:
				if self._main_above:
					if col_main:
						self._main_panel.on_touch_down(touch)
					elif col_side:
						self._side_panel.on_touch_down(touch)
				elif col_side:
					self._side_panel.on_touch_down(touch)
				elif col_main:
					self._main_panel.on_touch_down(touch)
				return False
		
		Animation.cancel_all(self)
		self._anim_init_progress = self._anim_progress
		self._touch = touch
		touch.ud['type'] = self.state
		touch.ud['panels_jiggled'] = False  # If user moved panels back
											# and forth, don't default
											# to close on touch release
		touch.grab(self)
		return True

	def on_touch_move(self, touch):
		if touch is self._touch:
			dx = touch.x - touch.ox
			self._anim_progress = max(0, min(self._anim_init_progress +
											(dx / self.side_panel_width), 1))
			if self._anim_progress < 0.975:
				touch.ud['panels_jiggled'] = True
		else:
			super().on_touch_move(touch)
			return

	def on_touch_up(self, touch):
		if touch is self._touch:
			self._touch = None
			init_state = touch.ud['type']
			touch.ungrab(self)
			jiggled = touch.ud['panels_jiggled']
			if init_state == 'open' and not jiggled:
				if self._anim_progress >= 0.975:
						self.anim_to_state('closed')
				else:
					self._anim_relax()
			else:
				self._anim_relax()
		else:
			super().on_touch_up(touch)
			return

	def _anim_relax(self):
		'''Animates to the open or closed position, depending on whether the
		current position is past self.min_dist_to_open.

		'''
		if self._anim_progress > self.min_dist_to_open:
			self.anim_to_state('open')
		else:
			self.anim_to_state('closed')

	def _choose_image(self, *args):
		'''Chooses which image to display as the main/side separator, based on
		_main_above.'''
		if self.separator_image:
			return self.separator_image
		if self._main_above:
			return 'kivygo/images/navigationdrawer_gradient_rtol.png'
		else:
			return 'kivygo/images/navigationdrawer_gradient_ltor.png'
